import json
import logging
import traceback

from qgis.gui import (QgsMapToolEmitPoint)  # pylint: disable=import-error
from qgis.core import Qgis  # Add this import for message bar

from ..logger import Logger
from ..logger import log
from ..street_smart import AbstractCommand

logger = Logger(__name__).get()

class ShowCycloramaCommand(AbstractCommand):
    """
    Class for Show Cyclorama Command
    """
    def __init__(self, iface, streetsmart):
        """
        Constructor
        """
        super().__init__(iface, streetsmart)
        super().set_disabled()
        self.streetsmart.buttonstate.atlas_wfs_added.connect(self.set_enabled)
        self.streetsmart.buttonstate.atlas_wfs_removed.connect(
            self.set_disabled)
        

        # Keep track of whether the tool is currently active
        self._tool_active = False
        # Create the map tool
        self._create_select_point_tool()

    def _create_select_point_tool(self):
        """
        Create and setup the select point map tool
        """
        # Ensure we're creating a new map tool each time
        self.select_point_tool = QgsMapToolEmitPoint(self.iface.mapCanvas())
        
        # Disconnect any existing connections to prevent multiple connections
        try:
            self.select_point_tool.canvasClicked.disconnect()
        except TypeError:
            # This is normal if no connections exist
            pass
        
        # Connect the canvas clicked signal
        self.select_point_tool.canvasClicked.connect(self.canvas_clicked)

    def callback(self):
        """
        Code to execute when command is clicked
        """
        try:
            if self.are_usersettings_available():
                # Recreate the select point tool to ensure it's fresh
                self._create_select_point_tool()
                
                # Set the map tool
                self.iface.mapCanvas().setMapTool(self.select_point_tool)
                
                # Mark the tool as active
                self._tool_active = True
                
                # Show a message in the interface
                self.iface.messageBar().pushMessage(
                    "Street Smart", 
                    "Select a point on the map", 
                    level=Qgis.Info
                )
            else:
                self.show_set_settings_message_box()
        except Exception as e:
            # Error handling
            self.iface.messageBar().pushMessage(
                "Street Smart", 
                f"Error setting up point selection: {e}", 
                level=Qgis.Critical
            )
            logger.error(f"Error in callback: {e}")
            logger.error(traceback.format_exc())


    @staticmethod
    def icon_path():
        """
        Path for the icon
        """
        return ':/plugins/street_smart/resources/mActionSelect.svg'

    def text(self):
        """
        Text for the command to show
        """
        return super().tr('Select Cyclorama point')

    def parent(self):
        """
        Parent
        """
        return self.iface.mainWindow()



    def canvas_clicked(self, point, button):
        """ Draw cone on the clicked point """
        try:
            # Ensure the tool is still active
            if not self._tool_active:
                self.iface.messageBar().pushMessage(
                    "Street Smart", 
                    "Please click 'Select Cyclorama Point' again", 
                    level=Qgis.Warning
                )
                return

            # Check if panorama viewer is visible
            if not self.streetsmart.panorama_viewer_visible:
                self.streetsmart.show_viewer()

            # Send point to viewer
            msg = self.__create_point_string(point)
            msg = "open|" + msg + '\n'
            self.streetsmart.sendToViewer(msg)

            # Mark viewer as open
            self.streetsmart.buttonstate.viewer_open = True

            # Show success message
            self.iface.messageBar().pushMessage(
                "Street Smart", 
                "Point selected successfully", 
                level=Qgis.Success
            )

            # Ensure the map tool remains active
            self.iface.mapCanvas().setMapTool(self.select_point_tool)

        except Exception as e:
            # Error handling
            self.iface.messageBar().pushMessage(
                "Street Smart", 
                f"Error selecting point: {e}", 
                level=Qgis.Critical
            )
            logger.error(f"Error in canvas_clicked: {e}")
            logger.error(traceback.format_exc())
            
            # Reset tool state if something goes wrong
            self._tool_active = False

    def __create_point_string(self, point) -> str:
            """Creates JSON string from point"""
            point_str = "{},{}".format(point.x(), point.y())
            crs = self.iface.mapCanvas().mapSettings().destinationCrs().authid()
            msg = {
                "point": point_str,
                "crs": crs
            }
            return json.dumps(msg).replace('\n', '')