""" Cone Handler module """
import math
import logging
from qgis.PyQt.QtCore import QObject, QEvent  # Add QObject import
from qgis.PyQt.QtGui import QColor  # pylint: disable=import-error
from qgis.core import (  # pylint: disable=import-error
    QgsPointXY, QgsGeometry, QgsProject, QgsCoordinateTransform,
    QgsCoordinateReferenceSystem, Qgis, QgsWkbTypes)
from qgis.gui import (  # pylint: disable=import-error
    QgsRubberBand)
try:
    from .settings import Settings
    from .logger import Logger
    from .logger import log
    from .geometry_utils import (transform_coordinate)
    from .utils import (convert_distance)
    from .street_smart import StreetSmart_Cone
except ImportError:
    from settings import Settings
    from logger import Logger
    from logger import log
    from geometry_utils import (transform_coordinate)
    from utils import (convert_distance)
    from street_smart import StreetSmart_Cone

CONE_BORDER_WIDTH = 3

logger = Logger(__name__).get()


class ConeHandler(QObject):  # Inherit from QObject
    """ 
    Cone Handler - Manages the visualization of the viewing cone in QGIS 
    that represents the panorama camera field of view.
    """

    def __init__(self, streetsmart, cone_height):
        """
        Initialize the cone handler
        
        Args:
            streetsmart: StreetSmart plugin instance
            cone_height: Initial cone height
        """
        super(ConeHandler, self).__init__()  # Initialize QObject parent class
        self.streetsmart = streetsmart
        self.iface = streetsmart.iface
        self.cone_height = cone_height
        self.cone = None
        # Store the last cone parameters for updating on zoom
        self._last_cone_params = None
        
        # Set up interaction for the cone
        self._setup_cone_interaction()
        
    def _setup_cone_interaction(self):
        """Set up interactions for the cone"""
        # Connect to map canvas events for cone manipulation
        canvas = self.iface.mapCanvas()
        
        # Install event filter for mouse events
        # This is the proper way to handle mouse events on the map canvas
        canvas.installEventFilter(self)
        
        # Track interaction state
        self._dragging = False
        self._drag_start_pos = None
        self._drag_start_orientation = None
        
    def eventFilter(self, watched_obj, event):
        """
        Event filter to handle map canvas events
        
        Args:
            watched_obj: The object being watched
            event: The event that occurred
            
        Returns:
            True if event was handled, False otherwise
        """
        from PyQt5.QtCore import QEvent
        
        # Only process events for the map canvas
        if watched_obj != self.iface.mapCanvas():
            return False
        
        # Handle different event types
        if event.type() == QEvent.MouseButtonPress:
            logger.debug("Mouse press event detected")
            result = self._on_mouse_press(event)
            return result  # Return True if we handled it
        elif event.type() == QEvent.MouseMove:
            if self._dragging:  # Only process if we're dragging
                result = self._on_mouse_move(event)
                return result  # Return True only if we're dragging
            return False
        elif event.type() == QEvent.MouseButtonRelease:
            if self._dragging:  # Only process if we were dragging
                result = self._on_mouse_release(event)
                return result  # Return True only if we were dragging
            return False
        elif event.type() == QEvent.Wheel:
            result = self._on_wheel_event(event)
            return result  # Return True if we handled it
            
        # For other events, let them propagate
        return False

    def handle(self, args):
        """ 
        Handle cone command from panorama viewer 
        
        Args:
            args: Command arguments string (comma-separated values) or None to remove the cone
        """
        # Remove existing cone if args is None
        if args is None:
            self.__remove_previous_cone()
            self._last_cone_params = None
            return

        # Parse arguments
        project_srs = self.iface.mapCanvas().mapSettings().destinationCrs()
        self.cone = StreetSmart_Cone(None, project_srs)
        
        # Handle both standard cone command and zoom change command
        if ',' in args:
            params = args.split(',')
            
            # Check the number of parameters to determine command type
            if len(params) == 10:
                # Standard cone command: r_id, x, y, z, srs, rot, r, g, b, a
                r_id, x, y, z, srs, rot, r, g, b, a = params
                
                # Store parameters for later use with zoom updates
                self._last_cone_params = {
                    'r_id': r_id,
                    'x': float(x),
                    'y': float(y),
                    'z': z,
                    'srs': srs,
                    'rot': rot,
                    'color': QColor(int(r), int(g), int(b), int(float(a))),
                    'zoom_factor': 1.0  # Default zoom factor
                }
                
                # Transform point to map projection
                transformed_point = self._transform_cone_point(
                    float(x), float(y), srs, project_srs)
                
                # Create and draw cone
                self.create_and_draw_cone(
                    transformed_point, 
                    rot,
                    self._last_cone_params['color'],
                    zoom_factor=1.0  # Default zoom factor
                )
                
                # Center map on the cone location
                self.center_map(transformed_point)
                
            elif len(params) == 11:
                # Extended cone command with zoom: r_id, x, y, z, srs, rot, r, g, b, a, zoom_factor
                r_id, x, y, z, srs, rot, r, g, b, a, zoom_factor = params
                
                try:
                    zoom_factor = float(zoom_factor)
                except ValueError:
                    logger.error(f"Invalid zoom factor in combined data: {zoom_factor}")
                    zoom_factor = 1.0
                
                # Store parameters for later use
                self._last_cone_params = {
                    'r_id': r_id,
                    'x': float(x),
                    'y': float(y),
                    'z': z,
                    'srs': srs,
                    'rot': rot,
                    'color': QColor(int(r), int(g), int(b), int(float(a))),
                    'zoom_factor': zoom_factor
                }
                
                # Transform point to map projection
                transformed_point = self._transform_cone_point(
                    float(x), float(y), srs, project_srs)
                
                # Create and draw cone with zoom factor
                self.create_and_draw_cone(
                    transformed_point, 
                    rot,
                    self._last_cone_params['color'],
                    zoom_factor=zoom_factor
                )
                
                # Center map on the cone location
                self.center_map(transformed_point)
                
            elif len(params) == 2 and params[0] == 'zoom':
                # Separate zoom command: zoom, zoom_factor
                zoom_factor = float(params[1])
                self.update_cone_for_zoom(zoom_factor)
        else:
            # Single zoom value
            try:
                zoom_factor = float(args)
                self.update_cone_for_zoom(zoom_factor)
            except ValueError:
                logger.error(f"Invalid zoom factor: {args}")

    def update_cone_for_zoom(self, zoom_factor):
        """
        Update the cone based on the zoom factor from the panorama viewer
        
        Args:
            zoom_factor: Zoom factor (1.0 is default, <1.0 is zoomed out, >1.0 is zoomed in)
        """
        if not self._last_cone_params:
            logger.warning("Cannot update cone: no previous cone parameters available")
            return
            
        # Get the stored cone parameters
        params = self._last_cone_params
        
        # Update zoom factor in stored parameters
        params['zoom_factor'] = zoom_factor
        
        # Transform the point to map projection
        project_srs = self.iface.mapCanvas().mapSettings().destinationCrs()
        transformed_point = self._transform_cone_point(
            params['x'], params['y'], params['srs'], project_srs)
        
        # Draw the updated cone
        self.create_and_draw_cone(
            transformed_point,
            params['rot'],
            params['color'],
            zoom_factor
        )

    def center_map(self, point):
        """
        Center the map on the given point
        
        Args:
            point: QgsPointXY to center on
        """
        # Convert the QgsPointXY to a QgsGeometry object
        geometry_feature = QgsGeometry.fromPointXY(point)
        
        # Get the center point of the feature's bounding box
        center_point = geometry_feature.boundingBox().center()
        
        # Set the center of the map canvas to the center point
        self.iface.mapCanvas().setCenter(center_point)

    @log(logging.DEBUG, print_args=True, print_return=True)
    def _transform_cone_point(self, x, y, srs, to_srs):
        """
        Transform coordinate to mapCanvas CRS
        
        Args:
            x: X coordinate
            y: Y coordinate
            srs: Source CRS
            to_srs: Target CRS
            
        Returns:
            Transformed QgsPointXY
        """
        return transform_coordinate(x, y, srs, to_srs)

    def _create_cone_polyline(self, point, zoom_factor=1.0):
        """ 
        Create the cone polygon from a top point 
        
        Args:
            point: QgsPointXY at the top of the cone
            zoom_factor: Zoom factor to adjust the cone angle (default: 1.0)
            
        Returns:
            List of QgsPointXY lists defining the cone polygon
        """
        # Convert cone height to canvas units
        canvas_units = self.iface.mapCanvas().mapUnits()
        height = convert_distance(self.cone_height, 0, canvas_units)
        
        # Get cone angle from settings
        base_angle = Settings.getInstance().get_cone_angle()
        
        # Adjust angle based on zoom factor
        # When zoomed in (zoom_factor > 1.0), the angle gets narrower
        # When zoomed out (zoom_factor < 1.0), the angle gets wider
        # Use inverse relationship: angle ∝ 1/zoom
        adjusted_angle = base_angle / zoom_factor
        
        # Cap the angle at 89 degrees to prevent extreme values
        if adjusted_angle > 89:
            adjusted_angle = 89
            
        # Get cone top coordinates
        cone_top_x = point.x()
        cone_top_y = point.y()

        # Calculate cone dimensions
        h_fov = math.radians(adjusted_angle)
        scale = self.iface.mapCanvas().scale()
        tana = math.tan(h_fov)

        real_height = height * scale
        real_width = tana * real_height

        # Create polygon points
        points = [[
            QgsPointXY(cone_top_x, cone_top_y),
            QgsPointXY(cone_top_x - real_width, cone_top_y + real_height),
            QgsPointXY(cone_top_x + real_width, cone_top_y + real_height)
        ]]
        return points

    def __remove_previous_cone(self):
        """ Remove previous cone from map canvas """
        previous_cone = self.streetsmart.buttonstate.previous_cone
        if previous_cone is not None and previous_cone.cone is not None:
            self.iface.mapCanvas().scene().removeItem(previous_cone.cone)

    def create_and_draw_cone(self, point, orientation, color, zoom_factor=1.0):
        """ 
        Draw cone on canvas 
        
        Args:
            point: QgsPointXY at the top of the cone
            orientation: Rotation angle in degrees
            color: QColor for the cone
            zoom_factor: Zoom factor to adjust cone angle (default: 1.0)
        """
        # Remove previous cone
        self.__remove_previous_cone()
        
        # Create cone polygon with zoom factor adjustment
        points = self._create_cone_polyline(point, zoom_factor)
        
        # Create geometry
        cone_geometry = QgsGeometry.fromPolygonXY(points)
        
        # Rotate cone
        try:
            rotation_angle = float(orientation)
            cone_geometry.rotate(rotation_angle, QgsPointXY(point.x(), point.y()))
        except (ValueError, TypeError) as e:
            logger.error(f"Error rotating cone: {str(e)}")
            rotation_angle = 0.0
        
        # Create and draw cone
        cone = QgsRubberBand(self.iface.mapCanvas(), QgsWkbTypes.PolygonGeometry)
        cone.setToGeometry(cone_geometry, None)
        cone.setColor(color)
        cone.setWidth(CONE_BORDER_WIDTH)
        
        # Make the cone more visible by increasing opacity if needed
        if color.alpha() < 128:
            new_color = QColor(color)
            new_color.setAlpha(128)
            cone.setColor(new_color)

        # Store the cone in the buttonstate
        self.streetsmart.buttonstate.previous_cone = StreetSmart_Cone(cone, self.cone.srs)
        
        # Log cone creation
        logger.debug(f"Created cone at {point.x()}, {point.y()} with orientation {rotation_angle}")
        
        # Refresh the map canvas
        self.iface.mapCanvas().refresh()

    def _on_mouse_press(self, event):
        """
        Handle mouse press events on the canvas
        
        Args:
            event: Mouse event
        """
        # Check if we have a cone to interact with
        if not self._last_cone_params or not self.streetsmart.buttonstate.previous_cone:
            return False
            
        # Get map coordinates from click
        point = self.iface.mapCanvas().getCoordinateTransform().toMapCoordinates(
            event.x(), event.y())
            
        # Check if the click is inside or near the cone
        cone = self.streetsmart.buttonstate.previous_cone.cone
        if cone and self._is_point_near_cone(point, cone):
            # Start dragging
            self._dragging = True
            self._drag_start_pos = point
            
            # Store current orientation
            if self._last_cone_params:
                self._drag_start_orientation = float(self._last_cone_params['rot'])
            
            return True  # We handled the event
            
        return False
            
    def _on_mouse_move(self, event):
        """
        Handle mouse move events on the canvas
        
        Args:
            event: Mouse event
            
        Returns:
            True if event was handled, False otherwise
        """
        if not self._dragging or not self._last_cone_params or not self._drag_start_pos:
            return False
            
        # Get map coordinates
        point = self.iface.mapCanvas().getCoordinateTransform().toMapCoordinates(
            event.x(), event.y())
            
        logger.debug(f"Mouse move while dragging: {point.x()}, {point.y()}")
        
        # Get cone center
        cone_x = self._last_cone_params['x']
        cone_y = self._last_cone_params['y']
        cone_point = self._transform_cone_point(
            cone_x, cone_y, 
            self._last_cone_params['srs'],
            self.iface.mapCanvas().mapSettings().destinationCrs()
        )
        
        # Calculate angle between points relative to cone center
        angle_start = self._calculate_angle(
            cone_point.x(), cone_point.y(),
            self._drag_start_pos.x(), self._drag_start_pos.y()
        )
        
        angle_current = self._calculate_angle(
            cone_point.x(), cone_point.y(),
            point.x(), point.y()
        )
        
        # Calculate angle difference
        angle_diff = angle_current - angle_start
        
        # Apply to baseline orientation
        new_orientation = (self._drag_start_orientation + angle_diff) % 360
        
        logger.debug(f"Calculated new orientation: {new_orientation}")
        
        # Update the cone
        self._update_cone_orientation(new_orientation)
        
        # Send the update to panorama viewer
        self._send_orientation_to_panorama(new_orientation)
        
        # We handled the event
        return True
        
    def _on_mouse_release(self, event):
        """
        Handle mouse release events on the canvas
        
        Args:
            event: Mouse event
            
        Returns:
            True if event was handled, False otherwise
        """
        if self._dragging:
            self._dragging = False
            self._drag_start_pos = None
            self._drag_start_orientation = None
            return True  # We handled the event
            
        return False
            
    def _on_wheel_event(self, event):
        """
        Handle mouse wheel events on the canvas
        
        Args:
            event: Wheel event
            
        Returns:
            True if event was handled, False otherwise
        """
        # Check if we have a cone to interact with
        if not self._last_cone_params or not self.streetsmart.buttonstate.previous_cone:
            return False
            
        # Get map coordinates from wheel position
        point = self.iface.mapCanvas().getCoordinateTransform().toMapCoordinates(
            event.x(), event.y())
            
        # Check if the wheel event is inside or near the cone
        cone = self.streetsmart.buttonstate.previous_cone.cone
        if cone and self._is_point_near_cone(point, cone):
            # Get current zoom factor
            current_zoom = self._last_cone_params.get('zoom_factor', 1.0)
            
            # Calculate zoom delta (120 is a standard wheel delta value)
            delta = event.angleDelta().y() / 120.0
            
            # Apply zoom change (15% per wheel step)
            new_zoom = max(0.1, current_zoom * (1.0 + (0.15 * delta)))
            
            # Update the cone
            self.update_cone_for_zoom(new_zoom)
            
            # Send the update to panorama viewer
            self._send_zoom_to_panorama(new_zoom)
            
            return True  # We handled the event
            
        return False
            
    def _is_point_near_cone(self, point, cone):
        """
        Check if a point is near the cone
        
        Args:
            point: QgsPointXY to check
            cone: QgsRubberBand cone
            
        Returns:
            True if point is near the cone
        """
        # Get cone geometry
        geometry = cone.asGeometry()
        if not geometry:
            logger.warning("Cannot get cone geometry")
            return False
            
        # Create a buffer around the geometry (20 pixels)
        map_units_per_pixel = self.iface.mapCanvas().mapUnitsPerPixel()
        buffer_size = map_units_per_pixel * 20  # Increased buffer for easier selection
        buffer_geometry = geometry.buffer(buffer_size, 5)
        
        # Debug logging
        logger.debug(f"Checking if point {point.x()},{point.y()} is near cone")
        
        # Check if point is inside buffer
        point_geom = QgsGeometry.fromPointXY(point)
        result = buffer_geometry.contains(point_geom) or buffer_geometry.intersects(point_geom)
        
        if result:
            logger.debug("Point is near cone")
        else:
            logger.debug("Point is NOT near cone")
            
        return result
        
    def _calculate_angle(self, cx, cy, px, py):
        """
        Calculate angle in degrees between points
        
        Args:
            cx: Center X
            cy: Center Y
            px: Point X
            py: Point Y
            
        Returns:
            Angle in degrees
        """
        dx = px - cx
        dy = py - cy
        
        # Prevent division by zero
        if abs(dx) < 1e-10 and abs(dy) < 1e-10:
            return 0
            
        angle = math.degrees(math.atan2(dy, dx))
        
        # Convert to 0-360 range
        return (angle + 360) % 360
        
    def _update_cone_orientation(self, orientation):
        """
        Update cone orientation without sending to panorama viewer
        
        Args:
            orientation: New orientation in degrees
        """
        if not self._last_cone_params:
            return
            
        # Update stored parameters
        self._last_cone_params['rot'] = str(orientation)
        
        # Get current parameters
        params = self._last_cone_params
        project_srs = self.iface.mapCanvas().mapSettings().destinationCrs()
        
        # Transform the point to map projection
        transformed_point = self._transform_cone_point(
            params['x'], params['y'], params['srs'], project_srs)
        
        # Create and draw cone
        self.create_and_draw_cone(
            transformed_point,
            str(orientation),
            params['color'],
            params.get('zoom_factor', 1.0)
        )
        
    def _send_orientation_to_panorama(self, orientation):
        """
        Send orientation update to panorama viewer
        
        Args:
            orientation: New orientation in degrees
        """
        # Format: "rotate|yaw"
        cmd = f"sendmessage|qgis:rotate|{orientation}"
        logger.debug(f"Sending orientation command to panorama viewer: {cmd}")
        self.streetsmart.sendToViewer(cmd)
        
    def _send_zoom_to_panorama(self, zoom_factor):
        """
        Send zoom update to panorama viewer
        
        Args:
            zoom_factor: New zoom factor
        """
        # Format: "zoom|factor"
        cmd = f"sendmessage|qgis:zoom|{zoom_factor}"
        logger.debug(f"Sending zoom command to panorama viewer: {cmd}")
        self.streetsmart.sendToViewer(cmd)