/**
 * Send combined cone and zoom update to QGIS
 * @param {viewer} panoramaViewer Viewer to get data from
 */
function sendCombinedUpdate(panoramaViewer) {
    // Get current zoom factor
    const currentZoom = getZoomFactor(panoramaViewer);
    
    // Get cone string (position, orientation, color)
    const cone = coneString(panoramaViewer);
    
    // Combine information: use the regular cone string and append zoom factor
    const combinedData = cone + "," + currentZoom;
    
    console.log("Sending combined update with zoom:", currentZoom);
    
    // Send the combined data to QGIS
    js_cone(combinedData);
    
    // Update stored zoom level
    window.lastZoomLevel = currentZoom;
}/*jshint esversion: 6 */

/**
 * Calls python to indicate that the browser is ready to receive commands
 */
function apiInitialized() {
    js_initialized();
    console.log('Api: init: success!');
    // Hide the loading screen and show the main content
    document.getElementById('loadingScreen').style.display = 'none';
    document.getElementById('streetsmartApi').style.display = 'block';
}

function loadError(oError) {
    console.log('Api: init: failed!', oError.target.src);
    // Optionally, display a different message or UI element to the user indicating the load failure
    document.getElementById('loadingScreen').innerHTML = "Failed to load resources. Please try again.";
}

/**
 * Initialize Street Smart API
 */
function initApi(userSettings, addressSettings, configSettings) {
    console.log("initApi");
    console.log(userSettings);
    console.log(addressSettings);
    console.log(configSettings);
    load_streetsmart_api();

    /**
     * Loads streetsmart API script
     * http://www.jspatterns.com/the-ridiculous-case-of-adding-a-script-element/
     */
    function load_streetsmart_api() {
        var js = document.createElement('script');
        js.onerror = loadError;
        js.async = false;
        js.src = configSettings.streetsmartApiUrl;
        js.onreadystatechange= initialize;
        js.onload = initialize;
        var first = document.getElementsByTagName('script')[0];
        first.parentNode.insertBefore(js, first);
        console.log(js);
    }
    
    /*
     * Initializes the streetsmart API once the script is loaded.
     */
    function initialize() {
        StreetSmartApi.init({
            targetElement: document.getElementById('streetsmartApi'),
            username: userSettings.username,
            password: userSettings.password,
            //loginOauth: true,
            //loginOauthSilentOnly: false,
            //clientId: 'DAC6C8E5-77AB-4F04-AFA5-D2A94DE6713F',
            apiKey: userSettings.apiKey,
            srs: configSettings.srs,
            locale: configSettings.locale,
            configurationUrl: configSettings.configUrl,
            addressSettings:
                {
                locale: addressSettings.locale,
                database: addressSettings.database
                }
        }).then(
            apiInitialized,
            function(err) {
                console.log('Api: init: failed. Error: ', err);
                alert('Api Init Failed!');
            }
        );
    }
}

/**
 * Creates the cone string
 * @param {viewer} panoramaViewer Viewer for which the cone is calculated
 */
function coneString(panoramaViewer) {
        rotation = panoramaViewer.getOrientation().yaw;
        x = panoramaViewer.getRecording().xyz[0];
        y = panoramaViewer.getRecording().xyz[1];
        z = panoramaViewer.getRecording().xyz[2];
        id = panoramaViewer.getRecording().id;
        srs = panoramaViewer.getRecording().srs;
        r = panoramaViewer.getViewerColor()[0];
        g = panoramaViewer.getViewerColor()[1];
        b = panoramaViewer.getViewerColor()[2];
        a = panoramaViewer.getViewerColor()[3] * 255;

        return [id, x, y, z, srs, rotation, r, g, b, a].toString();
}

// Store the last zoom level to detect changes
window.lastZoomLevel = 1.0; // Start with default zoom
window.defaultHFov = null;

/**
 * Gets the current zoom factor from the panorama viewer
 * @param {viewer} panoramaViewer Viewer to get zoom from
 * @returns {number} Zoom factor where 1.0 is the default/initial zoom
 */
function getZoomFactor(panoramaViewer) {
    // The API might provide different ways to access zoom level
    // Here are some possible approaches:
    
    // Option 1: If the API provides a direct zoom level or scale
    if (panoramaViewer.getZoom) {
        return panoramaViewer.getZoom();
    }
    
    // Option 2: Calculate from field of view (FOV)
    // Most panorama viewers use horizontal or vertical FOV
    // A smaller FOV means more zoom
    if (panoramaViewer.getOrientation && panoramaViewer.getOrientation().hFov) {
        // Get the current horizontal FOV
        const currentHFov = panoramaViewer.getOrientation().hFov;
        
        // Get the default/initial FOV (you might need to store this when viewer is first loaded)
        // If not available, use a typical default (often around 60-90 degrees)
        const defaultHFov = window.defaultHFov || 70;
        
        // Calculate zoom factor (inverse relationship - smaller FOV = higher zoom)
        return defaultHFov / currentHFov;
    }
    
    // Option 3: If zoom info is available in camera parameters
    if (panoramaViewer.getCameraParameters) {
        const params = panoramaViewer.getCameraParameters();
        if (params && params.zoom) {
            return params.zoom;
        }
    }
    
    // Fallback - return the last known zoom level or default
    return window.lastZoomLevel || 1.0;
}

/**
 * Send zoom update to QGIS
 * @param {viewer} panoramaViewer Viewer to get zoom from
 */
function updateZoom(panoramaViewer) {
    const currentZoom = getZoomFactor(panoramaViewer);
    
    // Only send updates when zoom actually changes
    // Use a small threshold to avoid sending too many updates for tiny changes
    if (window.lastZoomLevel === null || 
        Math.abs(currentZoom - window.lastZoomLevel) > 0.05) {
        
        console.log("Zoom changed to: " + currentZoom);
        
        // Send the zoom update to QGIS
        js_cone("zoom," + currentZoom);
        
        // Update stored zoom level
        window.lastZoomLevel = currentZoom;
    }
}

/**
 * Initialize default field of view value for zoom calculations
 * @param {viewer} panoramaViewer The panorama viewer
 */
function initDefaultFov(panoramaViewer) {
    // Store the initial FOV as the default/reference point
    if (panoramaViewer.getOrientation && panoramaViewer.getOrientation().hFov) {
        window.defaultHFov = panoramaViewer.getOrientation().hFov;
        console.log("Set default hFov to: " + window.defaultHFov);
    }
}

/**
 * Apply a zoom level to the panorama viewer
 * @param {viewer} panoramaViewer The panorama viewer
 * @param {number} zoomFactor Zoom factor to apply
 */
function applyZoomToViewer(panoramaViewer, zoomFactor) {
    console.log("Applying zoom factor:", zoomFactor);
    
    // Method 1: If viewer has a direct setZoom method
    if (panoramaViewer.setZoom) {
        panoramaViewer.setZoom(zoomFactor);
        return;
    }
    
    // Method 2: If viewer uses FOV for zoom
    if (panoramaViewer.setOrientation && window.defaultHFov) {
        // Calculate FOV from zoom factor (inverse relationship)
        const targetFov = window.defaultHFov / zoomFactor;
        
        // Get current orientation
        const currentOrientation = panoramaViewer.getOrientation();
        
        // Create new orientation with adjusted FOV
        const newOrientation = {
            ...currentOrientation,
            hFov: targetFov
        };
        
        // Apply the new orientation
        panoramaViewer.setOrientation(newOrientation);
        return;
    }
    
    // Method 3: If there's a camera control method
    if (panoramaViewer.setCameraParameters) {
        panoramaViewer.setCameraParameters({ zoom: zoomFactor });
        return;
    }
    
    console.log("No suitable method found to apply zoom");
}

/**
 * Handle commands received from QGIS
 * @param {string} command Command string
 */
function handleQgisCommands(command) {
    if (!command || !window.panoramaViewer) return;
    
    console.log("Received command from QGIS:", command);
    
    // Parse the command
    const parts = command.split("|");
    
    if (parts.length < 2) return;
    
    const commandType = parts[0];
    const commandData = parts[1];
    
    switch (commandType) {
        case "rotate":
            // Format: rotate|yaw
            const yaw = parseFloat(commandData);
            if (!isNaN(yaw)) {
                rotatePanorama(yaw);
            }
            break;
            
        case "zoom":
            // Format: zoom|factor
            const zoomFactor = parseFloat(commandData);
            if (!isNaN(zoomFactor)) {
                applyZoomToViewer(window.panoramaViewer, zoomFactor);
                window.lastZoomLevel = zoomFactor;
            }
            break;
            
        case "orientation":
            // Format: orientation|yaw,pitch,roll,zoom
            const params = commandData.split(",");
            if (params.length >= 3) {
                const yaw = parseFloat(params[0]) || 0;
                const pitch = parseFloat(params[1]) || 0;
                const roll = parseFloat(params[2]) || 0;
                const zoom = params.length > 3 ? parseFloat(params[3]) : window.lastZoomLevel;
                
                setFullOrientation(yaw, pitch, roll, zoom);
            }
            break;
    }
}

/**
 * Rotate the panorama to a specific yaw angle
 * @param {number} yaw Yaw angle in degrees
 */
function rotatePanorama(yaw) {
    if (!window.panoramaViewer) return;
    
    const currentOrientation = window.panoramaViewer.getOrientation();
    const newOrientation = {
        ...currentOrientation,
        yaw: yaw
    };
    
    window.panoramaViewer.setOrientation(newOrientation);
}

/**
 * Set full orientation including zoom
 * @param {number} yaw Yaw angle in degrees
 * @param {number} pitch Pitch angle in degrees
 * @param {number} roll Roll angle in degrees
 * @param {number} zoom Zoom factor
 */
function setFullOrientation(yaw, pitch, roll, zoom) {
    if (!window.panoramaViewer) return;
    
    // Get current orientation to preserve any values we're not changing
    const currentOrientation = window.panoramaViewer.getOrientation();
    
    // Calculate FOV if we're using that for zoom
    let hFov = null;
    if (window.defaultHFov && zoom) {
        hFov = window.defaultHFov / zoom;
    }
    
    // Create new orientation object
    const newOrientation = {
        ...currentOrientation,
        yaw: yaw,
        pitch: pitch,
        roll: roll
    };
    
    // Add FOV if we calculated it
    if (hFov !== null) {
        newOrientation.hFov = hFov;
    }
    
    // Apply the new orientation
    window.panoramaViewer.setOrientation(newOrientation);
    
    // Update our stored zoom level
    if (zoom) {
        window.lastZoomLevel = zoom;
    }
}

// Set up a global function to receive commands from QGIS
window.receiveQgisCommand = handleQgisCommands;

/**
 * Add eventhandlers to panoramaViewer
 */
function initEvents() {
    console.log("Set Measurement event");
    
    // Set up measurement changed event
    StreetSmartApi.on(StreetSmartApi.Events.measurement.MEASUREMENT_CHANGED, function(e){
        js_measure(JSON.stringify(e.detail.activeMeasurement));
    });
    
    // Store the initial field of view for zoom calculations
    initDefaultFov(window.panoramaViewer);
    
    // Set up panorama viewer events
    window.panoramaViewer
        .on(StreetSmartApi.Events.panoramaViewer.IMAGE_CHANGE, function(e){
            console.log("Image Changed", StreetSmartApi.Events.panoramaViewer);
            console.log(e);
            
            // On image change, preserve the last zoom factor if possible
            const currentZoom = window.lastZoomLevel || 1.0;
            
            // Reinitialize default FOV for this new image
            initDefaultFov(window.panoramaViewer);
            
            // Send update with preserved zoom
            const cone = coneString(window.panoramaViewer);
            const combinedData = cone + "," + currentZoom;
            js_cone(combinedData);
            
            // If there's a way to apply the zoom level to the new image, do it here
            applyZoomToViewer(window.panoramaViewer, currentZoom);
        })
        .on(StreetSmartApi.Events.panoramaViewer.VIEW_CHANGE, function(e){
            // For any view change (click, drag, etc.), always send full information
            // including current zoom level - don't let it reset
            sendCombinedUpdate(window.panoramaViewer);
        })
        .on(StreetSmartApi.Events.panoramaViewer.VIEW_LOAD_END, function(e){
            // When a new view loads, initialize values but maintain zoom if possible
            initDefaultFov(window.panoramaViewer);
            sendCombinedUpdate(window.panoramaViewer);
        })
        .on(StreetSmartApi.Events.panoramaViewer.VIEW_LOAD_START, function(e){
            console.log("View load start Changed");
            console.log(e);
        })
        .on(StreetSmartApi.Events.panoramaViewer.TIME_TRAVEL_CHANGE, function(e){
            console.log("Time travel change");
            console.log(e);
            
            // Time travel should maintain the zoom level
            sendCombinedUpdate(window.panoramaViewer);
        });
        
    // If the API provides a specific zoom change event, use it
    if (StreetSmartApi.Events.panoramaViewer.ZOOM_CHANGE) {
        window.panoramaViewer.on(StreetSmartApi.Events.panoramaViewer.ZOOM_CHANGE, function(e) {
            // On explicit zoom change, send complete information
            sendCombinedUpdate(window.panoramaViewer);
        });
    }
    
    // Add event listeners for mouse wheel and touch pinch zoom
    // These might help catch zoom events that aren't firing through the API
    const apiElement = document.getElementById('streetsmartApi');
    
    if (apiElement) {
        // Listen for wheel events (mouse wheel zoom)
        apiElement.addEventListener('wheel', function(e) {
            // Debounce the wheel events to avoid too many updates
            if (window.wheelTimer) {
                clearTimeout(window.wheelTimer);
            }
            
            window.wheelTimer = setTimeout(function() {
                sendCombinedUpdate(window.panoramaViewer);
            }, 150); // Small delay to avoid flooding with events
        }, { passive: true });
    }
}

/**
 * Add window loaded eventhandler
 */
window.addEventListener('load', function() {
    console.log("it's loaded!");
    js_window_loaded();
});

/*
 * Get a measurement
 *
 * Function is called form QGIS (obsolete)
 */
function getMeasure() {
    console.log("getMeasure called");
    activeMeasure = StreetSmartApi.getActiveMeasurement(window.panoramaViewer);
    console.log(activeMeasure);
}

/*
 * Starts a measurement with the corresponding geometry type
 *
 * Function is called form QGIS
 */
function startMeasure(geometryType) {
    measurementtype = '';
    if (geometryType == 'point') {
        measurementtype = StreetSmartApi.MeasurementGeometryType.POINT;
    } else if (geometryType == 'polyline') {
        measurementtype = StreetSmartApi.MeasurementGeometryType.LINESTRING;
    } else if (geometryType == 'polygon') {
        measurementtype = StreetSmartApi.MeasurementGeometryType.POLYGON;
    }

    if (measurementtype == '') {
        StreetSmartApi.startMeasurementMode(window.panoramaViewer);
    } else {
        StreetSmartApi.startMeasurementMode(window.panoramaViewer,
            {geometry:measurementtype});
    }
}

/*
 * Stops a measurement
 *
 * Function is called form QGIS
 */
function stopMeasure() {
    StreetSmartApi.stopMeasurementMode();
}

/**
 * Opens an image in a panorama viewer
 * @param {string} point Point to open Cyclorama
 *
 * Function is called from QGIS
 */
function open(point) {
    point = JSON.parse(point);
    StreetSmartApi.open(point.point, {
        viewerType: [StreetSmartApi.ViewerType.PANORAMA],
        srs: point.crs,
        panoramaViewer: { replace: true, },
    }).then(
        function(result) {
            if (result && result[0]) {
                for (let i = 0; i < result.length; i++) {
                    console.log('Opened a panorama viewer through API!', result[0]);
                    if(result[i].getType() === StreetSmartApi.ViewerType.PANORAMA) {
                        window.panoramaViewer = result[i];
                        // Reset zoom tracking when opening a new panorama
                        window.lastZoomLevel = null;
                        initEvents();
                    }
                }
            }
        }
    ).catch(
        function(reason) {
            console.log('Error opening panorama viewer: ' + reason);
            alert('Error opening panorama viewer: ' + reason);
        }
    );
}

// Overlay layers added to the viewer
window.overlayLayers = [];

/**
 * Remove overlays added to panorama viewer
 *
 * Function is called from QGIS
 */
function removeOverlay() {
    len = window.overlayLayers.length;
    for (i = 0; i < len; i++) {
        StreetSmartApi.removeOverlay(window.overlayLayers[i].id);
    }
}

/**
 * Add overlay to panorama viewer
 * @param {string} geojson Geometries to shown
 * @param {string} name Name of the layer
 * @param {string} srs Coordinate System Reference
 * @param {string} sldText SLD with which the geometries are shown
 * @param {string} color Color of the layer
 *
 * Function is called from QGIS
 */
function addOverlay(geojson, name, srs, sldText, color) {
    console.log("Try to add overlay");
    console.log(geojson);
    console.log(name);
    console.log(srs);
    console.log(sldText);
    console.log(color);

    options = {
        "geojson": JSON.parse(geojson),
        "visible": true,
        "id": '',
    };
    if (name !== undefined && name !== "") {
        options.name = name;
    }
    if (srs !== undefined && srs !== "") {
        options.sourceSrs = srs;
    }
    if (sldText !== undefined && sldText !== "") {
        options.sldXMLtext = sldText;
    }
    if (color !== undefined && color !== "") {
        options.color = color;
    }

    console.log(options);
    try {
        window.overlayLayers.push(StreetSmartApi.addOverlay(options));
    } catch (e) {
        alert(e);
    }
}