# -*- coding: utf-8 -*-
"""StreetSmart

This module is the main entrance for the QGIS plugin. The functions and classes
in this module make it possible to communicate between QGIS, the plugin
commands and the panorama viewer.

This module has been refactored to improve:
1. Code organization and modularity
2. Error handling and logging
3. Dependency management
4. Performance and resource usage
5. Thread safety
"""
import importlib
import inspect
import logging
import os
import os.path
import queue
import signal
import socket
import subprocess
import sys
import time
import typing
from abc import abstractmethod, ABCMeta
from threading import Thread
from pathlib import Path
from dataclasses import dataclass

from qgis.PyQt.QtCore import (
    QTimer, QSettings, QTranslator, QCoreApplication, QObject, pyqtSignal
)
from qgis.core import (
    QgsVectorLayer, QgsExpressionContextUtils, QgsProject, Qgis
)
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction, QMessageBox
from PyQt5.QtWidgets import QInputDialog, QVBoxLayout, QWidget, QCheckBox, QDialog, QLabel, QDialogButtonBox

# Local imports - use relative imports consistently
from .install import (
    install_cefpython3, check_DLLs, copy_missen_DLLS, 
    return_Qgis_bin_path, check_Cefpython_installation, 
    configure_paths, check_qgis_version
)
from .checkboxDLLs import CheckboxDialog
from .resources import *
from .command_getter import CommandGetter
from .logger import Logger, log, set_socket_port
from .settings import Settings
from .controller import initialize_servers
from .utils import start_python_script, convert_meters_to_mapunits
from .dependency_manager import (
    install_required_packages, 
    check_python_version,
    REQUIRED_PACKAGES
)

# Configure logger
logger = Logger(__name__).get()

# Constants
HOST = "localhost"
COMMANDSDIRECTORY = "commands"
# Development mode toggle
DEBUG_MODE = False


class AbstractCommand(metaclass=ABCMeta):
    """Definition of the interface a command must have."""
    
    def __init__(self, iface, streetsmart):
        """
        Constructor
        
        Args:
            iface: QGIS interface
            streetsmart: StreetSmart plugin instance
        """
        self.iface = iface
        self.__streetsmart = streetsmart
        self.__enabled = True

    @property
    def action(self):
        """Returns the Qt action for this command."""
        return self.streetsmart.get_action_for_command(self)

    @property
    def streetsmart(self):
        """Returns the Street Smart reference."""
        return self.__streetsmart

    @property
    def enabled(self):
        """Return true if the command is enabled."""
        return self.__enabled

    @enabled.setter
    def enabled(self, value):
        self.__enabled = value

    def __set_enabled(self, enabled):
        """Sets the enabled flag for a command."""
        action = self.streetsmart.get_action_for_command(self)
        if action:
            action.setEnabled(enabled)
        self.__enabled = enabled
        
    def set_enabled(self):
        """Sets a command to enabled."""
        self.__set_enabled(True)
        
    def set_disabled(self):
        """Sets a command to disabled."""
        self.__set_enabled(False)

    def show_set_settings_message_box(self):
        """Show message to set username and password."""
        title = self.tr('Add Atlas WFS')
        message = self.tr('Please set username and password first in the settings dialog box')
        QMessageBox.information(None, title, message, QMessageBox.Ok)

    @staticmethod
    @log(logging.DEBUG, print_return=True)
    def are_usersettings_available():
        """Returns true if username and password are set"""
        return bool(Settings.getInstance().settings[Settings.PASSWORD].value)

    @staticmethod
    def tr(message):
        """Returns the translated message"""
        return QCoreApplication.translate('StreetSmart', message)

    @staticmethod
    @abstractmethod
    def icon_path():
        """Returns the path of the icon to shown on the toolbar."""

    @abstractmethod
    def text(self):
        """Returns the tooltip text to be shown when hoovering the toolbar."""

    @abstractmethod
    def parent(self):
        """Parent of the action."""

    @abstractmethod
    def callback(self):
        """Method to be executed when the action is clicked."""


@dataclass
class StreetSmart_Cone:
    """A streetsmart cone data structure"""
    cone: typing.Any
    srs: str


class ButtonStateSubject(QObject):
    """Manages plugin state and emits signals on state changes."""
    
    # Define signals
    panorama_viewer_opened = pyqtSignal()
    panorama_viewer_closed = pyqtSignal()
    atlas_wfs_added = pyqtSignal()
    atlas_wfs_removed = pyqtSignal()
    new_project_created = pyqtSignal()
    project_read = pyqtSignal()
    map_canvas_refreshed = pyqtSignal()
    map_scale_changed = pyqtSignal()
    cone_moved = pyqtSignal()
    panorama_viewer_server_port_set = pyqtSignal()

    def __init__(self, iface, streetsmart):
        """
        Constructor
        
        Args:
            iface: Interface with QGIS
            streetsmart: StreetSmart plugin instance
        """
        super().__init__()
        self.__viewer_open = False
        self.__wfs_added = False
        self.__previous_cone = None
        self.iface = iface
        self.__panorama_viewer_server_port = 0
        self.__streetsmart = streetsmart

        # Connect signals
        self._connect_signals()
        
        # Check QGIS version compatibility
        self._check_compatibility()
        
        # Check required DLLs
        self._check_dependencies()

    def _connect_signals(self):
        """Connect all required QGIS signals"""
        self.iface.newProjectCreated.connect(self.on_project_created)
        self.iface.projectRead.connect(self.on_project_read)
        map_canvas = self.iface.mapCanvas()
        map_canvas.mapCanvasRefreshed.connect(self.on_map_canvas_refreshed)
        map_canvas.scaleChanged.connect(self.on_map_scale_changed)
        QgsProject.instance().layerWillBeRemoved.connect(self.on_layer_will_be_removed)

    def _check_compatibility(self):
        """Check if the plugin is compatible with the current QGIS version"""
        if not check_qgis_version():
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Warning)
            msg.setText("This version of the StreetSmart plugin is not compatible with your QGIS version.")
            msg.setWindowTitle("Compatibility Warning")
            msg.setStandardButtons(QMessageBox.Ok)
            msg.exec_()

    def _check_dependencies(self):
        """Check if required DLLs are available and install if needed"""
        if not check_Cefpython_installation():
            dlls_missing, path_checked, missing_files = check_DLLs()
            
            if dlls_missing:
                checkbox_dialog = CheckboxDialog(missing_files)
                result = checkbox_dialog.exec_()
                
                if result == QDialog.Accepted:
                    qgis_bin_path = return_Qgis_bin_path(path_checked)
                    for file in missing_files:
                        source_path = os.path.join(qgis_bin_path, file)
                        dest_path = os.path.join(path_checked, file)
                        copy_missen_DLLS(source_path, dest_path)
                        logger.info(f"Copied {file} to {path_checked}")

    @property
    def panorama_viewer_server_port(self):
        """Returns the port the panorama viewer is listening to."""
        return self.__panorama_viewer_server_port

    @panorama_viewer_server_port.setter
    def panorama_viewer_server_port(self, value):
        """Set the panorama viewer server port and emit signal if changed"""
        logger.debug("Set state port to %s", value)
        self.__panorama_viewer_server_port = value
        if value != 0:
            self.panorama_viewer_server_port_set.emit()

    @property
    def previous_cone(self):
        """Returns the cone which is currently shown on the mapCanvas"""
        return self.__previous_cone

    @staticmethod
    @log(logging.DEBUG, print_return=True)
    def get_rubberband_points(cone):
        """Returns the points of the given rubberband"""
        if cone:
            return [cone.getPoint(0, i) for i in range(cone.numberOfVertices())]
        return []

    def on_layer_will_be_removed(self, layer_id):
        """Handles the layer removed signal"""
        layer = QgsProject.instance().mapLayer(layer_id)
        default_layer_name = Settings.getInstance().getRecordingLayerName()
        if layer and layer.name() == default_layer_name:
            self.__streetsmart.hide_viewer()
            self.atlas_added = False

    @log(logging.DEBUG, print_args=True, print_return=True)
    def has_cone_moved(self, previous_cone, cone):
        """Returns True if the cone has moved."""
        if not previous_cone or not cone:
            return True
            
        previous_points = self.get_rubberband_points(previous_cone)
        current_points = self.get_rubberband_points(cone)

        # If the number of points differs, consider it moved
        if len(previous_points) != len(current_points):
            return True

        # Compare each point with a small tolerance
        epsilon = convert_meters_to_mapunits(1, self.iface.mapCanvas())
        for curr_point, prev_point in zip(current_points, previous_points):
            if not curr_point.compare(prev_point, epsilon=epsilon):
                return True
                
        return False

    @log(logging.DEBUG)
    def raise_if_cone_moved(self, previous_cone, cone):
        """Emit signal if the cone has moved"""
        if self.has_cone_moved(previous_cone, cone):
            self.cone_moved.emit()

    @previous_cone.setter
    def previous_cone(self, cone):
        """Set the previous cone and check if it has moved"""
        prev_cone = self.__previous_cone.cone if self.__previous_cone else None
        self.__previous_cone = cone
        self.raise_if_cone_moved(prev_cone, cone.cone if cone else None)

    def set_cone_srs(self, srs: str):
        """Set spatial reference for cone"""
        if self.__previous_cone:
            self.__previous_cone.srs = srs

    @property
    def viewer_open(self):
        """Returns true if the panorama viewer is visible"""
        return self.__viewer_open

    @viewer_open.setter
    def viewer_open(self, value):
        """Set the viewer open state and emit signals if changed"""
        if value == self.__viewer_open:
            return

        if value:
            logger.debug("Panorama Viewer Opened")
            self.panorama_viewer_opened.emit()
        else:
            logger.debug("Panorama Viewer Closed")
            self.panorama_viewer_closed.emit()

        self.__viewer_open = value

    @property
    def atlas_added(self):
        """Returns true if the Atlas WFS is added to the current project"""
        return self.__wfs_added

    @atlas_added.setter
    def atlas_added(self, value):
        """Set the atlas added state and emit signals if changed"""
        if value == self.__wfs_added:
            return
            
        if value:
            self.atlas_wfs_added.emit()
        else:
            self.atlas_wfs_removed.emit()

        self.__wfs_added = value

    def on_project_created(self):
        """Executes when a new project is created"""
        self.__streetsmart.hide_viewer()
        self.viewer_open = False
        self.new_project_created.emit()

    @staticmethod
    @log(level=logging.DEBUG, print_return=True)
    def get_legend_layer_names():
        """Returns a list with layer names"""
        layers = QgsProject.instance().mapLayers().values()
        return [l.name() for l in layers]

    @log(level=logging.DEBUG, print_return=True)
    def project_contains_atlas_layer(self):
        """Returns if the current project has the atlas layer in its legend."""
        layer_names = self.get_legend_layer_names()
        return Settings.getInstance().getRecordingLayerName() in layer_names

    def on_project_read(self):
        """Executes when an existing project is read"""
        self.__streetsmart.hide_viewer()
        if not self.project_contains_atlas_layer():
            self.viewer_open = False
            self.atlas_added = False
        else:
            self.atlas_added = True

        self.project_read.emit()

    def on_map_canvas_refreshed(self):
        """Executes when the mapCanvas is refreshed"""
        self.map_canvas_refreshed.emit()

    def on_map_scale_changed(self):
        """Executes when the mapScale is changed"""
        self.map_scale_changed.emit()


class MessageSender(Thread):
    """Thread to handle asynchronous message sending to the panorama viewer"""
    
    def __init__(self, state, msg):
        """
        Constructor
        
        Args:
            state: ButtonStateSubject instance
            msg: Message to send
        """
        super().__init__()
        self.__state = state
        self.__msg = msg
        self.daemon = True  # Make thread exit when main thread exits

    def run(self):
        """Thread execution logic"""
        logger.debug("MessageSender started")
        
        # Wait for port to be set (with timeout)
        timeout = 10  # seconds
        start_time = time.time()
        while self.__state.panorama_viewer_server_port == 0:
            if time.time() - start_time > timeout:
                logger.error("Timeout waiting for panorama viewer server port")
                return
            time.sleep(0.1)

        port = self.__state.panorama_viewer_server_port
        logger.debug(f"Got panorama viewer port {port}")
        
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                sock.settimeout(5)  # Set timeout for connect and send operations
                logger.debug(f"Sending '{self.__msg}' to port {port}")
                sock.connect((HOST, port))
                sock.sendall(bytes(self.__msg, "utf-8"))
                logger.debug(f"Message sent: {self.__msg}")
        except (socket.timeout, ConnectionRefusedError, OSError) as e:
            logger.error(f"Failed to send message: {str(e)}")


class StreetSmart:
    """QGIS Plugin Implementation for StreetSmart."""

    def __init__(self, iface):
        """
        Constructor.
        
        Args:
            iface: QGIS interface instance
        """
        logger.debug("Initializing StreetSmart Plugin")

        # Save reference to the QGIS interface
        self.iface = iface
        
        # Initialize plugin directory and locale
        self._init_locale()
        
        # Declare instance attributes
        self.actions = []
        self.commands = []
        self.proc = None
        self.__pv_server_port = 0
        self.menu = self.tr(u'&StreetSmart')
        self.toolbar = self.iface.addToolBar(u'Street Smart')
        self.toolbar.setObjectName(u'Street Smart')
        self.buttonstate = ButtonStateSubject(iface, self)
        self.command_queue = queue.Queue()
        self.command_handler = None
        self.log_port = None
        self.log_pid = None
        
        # Install required packages
        self._install_dependencies()

    def _init_locale(self):
        """Initialize locale for translations"""
        plugin_dir = os.path.dirname(__file__)
        locale = QSettings().value('locale/userLocale')[0:2]
        logger.debug(f"Locale: {locale}")
        
        locale_path = os.path.join(
            plugin_dir, 'i18n', f'StreetSmart_{locale}.qm')

        if os.path.exists(locale_path):
            logger.debug(f"Installing translator: {locale_path}")
            self.translator = QTranslator()
            self.translator.load(locale_path)
            QCoreApplication.installTranslator(self.translator)
        else:
            logger.debug(f"Translation file not found: {locale_path}")

    def _install_dependencies(self):
        """Install required packages if not available"""
        install_required_packages(REQUIRED_PACKAGES)

    @property
    def pv_server_port(self):
        """Returns the Panorama Viewer server port."""
        return self.__pv_server_port

    @pv_server_port.setter
    def pv_server_port(self, value):
        """Set the panorama viewer server port."""
        logger.debug(f"Setting port to {value}")
        self.buttonstate.panorama_viewer_server_port = value

    @property
    def panorama_viewer_visible(self):
        """Returns true if the panorama viewer is visible."""
        if self.proc is not None:
            # Check if subprocess is still running
            return self.proc.poll() is None
        return False

    def get_action_for_command(self, command):
        """
        Returns the QGIS action for the StreetSmart command.
        
        Args:
            command: Command instance
        
        Returns:
            QAction or None
        """
        if not self.commands:
            return None
            
        try:
            command_index = self.commands.index(command)
            return self.actions[command_index]
        except ValueError:
            return None

    @log(level=logging.DEBUG, print_args=True)
    def send_to_viewer(self, msg: str):
        """
        Sends message to the panorama viewer.
        
        Args:
            msg: Message to send
        """
        logger.debug(f"Sending message to PanoramaViewer: {msg}")
        sender = MessageSender(self.buttonstate, msg)
        sender.start()
        
    # Alias for backward compatibility with existing code
    sendToViewer = send_to_viewer

    @staticmethod
    def tr(message):
        """
        Get the translation for a string using Qt translation API.
        
        Args:
            message: String for translation
            
        Returns:
            Translated version of message
        """
        return QCoreApplication.translate('StreetSmart', message)

    def add_action(self, icon_path, text, callback, enabled_flag=True,
                  add_to_menu=True, add_to_toolbar=True, status_tip=None,
                  whats_this=None, parent=None):
        """
        Add a toolbar icon to the toolbar.
        
        Args:
            icon_path: Path to the icon
            text: Text for the action
            callback: Function to be called when action is triggered
            enabled_flag: Whether the action should be enabled by default
            add_to_menu: Whether to add the action to the menu
            add_to_toolbar: Whether to add the action to the toolbar
            status_tip: Optional text for status tip
            whats_this: Optional text for "what's this" help
            parent: Parent widget
            
        Returns:
            QAction that was created
        """
        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            self.toolbar.addAction(action)

        if add_to_menu:
            self.iface.addPluginToMenu(self.menu, action)

        self.actions.append(action)
        return action

    def initGui(self):
        """Creates the menu entries and toolbar icons inside the QGIS GUI."""
        # Initialize logging and servers
        log_pid, server_log_port = initialize_servers()
        logger.info(f"Initialized servers: log_pid={log_pid}, log_port={server_log_port}")
        
        self.log_port = server_log_port
        self.log_pid = log_pid
        set_socket_port(server_log_port)
        
        # Reinitialize logger with port
        Logger(__name__, port=server_log_port).get()
        logger.info(f"=== {__name__} initialized ===")

        # Show release notes if needed
        try:
            from .release_notes import show_release_notes
            show_release_notes(self.iface.mainWindow())
        except Exception as e:
            logger.error(f"Error showing release notes: {str(e)}")

        # Install commands
        self.install_commands()

    def current_layer_changed(self, new_current_layer):
        """
        Event handler when current layer changes
        
        Args:
            new_current_layer: The new current layer
        """
        is_vector_layer = isinstance(new_current_layer, QgsVectorLayer)
        if is_vector_layer and new_current_layer.isEditable():
            geometry_type = self.get_vectorlayer_feature_type(new_current_layer)
            geometry_type = geometry_type if geometry_type else 'point'  # Default
            self.sendToViewer(f"restartmeasure|{geometry_type}")
        else:
            self.sendToViewer("stopmeasure")

    @log(level=logging.DEBUG, print_return=True)
    def get_vectorlayer_feature_type(self, layer):
        """
        Returns the geometric type of features in the layer
        
        Args:
            layer: Vector layer
            
        Returns:
            String representation of geometry type
        """
        geometry_type = layer.geometryType()
        return self.__geometry_to_string(geometry_type)

    @staticmethod
    @log(level=logging.DEBUG, print_args=True, print_return=True)
    def __geometry_to_string(geometry_type: int):
        """
        Converts a geometry type to a readable string.
        
        Args:
            geometry_type: QGIS geometry type (integer)
            
        Returns:
            String representation of geometry type
        """
        geometry_map = {
            0: "point",
            1: "polyline",
            2: "polygon"
        }
        return geometry_map.get(geometry_type)

    @log(level=logging.DEBUG)
    def __add_actions(self, modules):
        """
        Add actions for commands
        
        Args:
            modules: List of command modules
        """
        toolbar_ordered_commands = [
            'AddAtlasWFSLayerCommand',
            'ShowCycloramaCommand',
            'AddOverlayCommand',
            'CreateMeasureCommand',
            'ManageSettingsCommand',
            'ShowHelpCommand',
        ]

        all_commands = {}
        for pkg_parent, command_module in modules:
            clsmembers = inspect.getmembers(command_module, inspect.isclass)
            for clsname, cls in clsmembers:
                if (cls.__module__.startswith(pkg_parent) and 
                    issubclass(cls, AbstractCommand)):
                    all_commands[clsname] = cls

        # First add commands in specified order
        for cmd in toolbar_ordered_commands:
            if cmd in all_commands:
                cls = all_commands[cmd]
                del all_commands[cmd]
                self.__create_action_from_class(cls, True)

        # Then add remaining commands
        for cls in all_commands.values():
            self.__create_action_from_class(cls, False)

        # Import QGISCommandHandler here to prevent cyclic imports
        from .qgis_command_handler import QGISCommandHandler
        self.command_handler = QGISCommandHandler(self, self.command_queue)

    def __create_action_from_class(self, command_class, add_to_toolbar):
        """
        Create an action from a command class
        
        Args:
            command_class: Command class to instantiate
            add_to_toolbar: Whether to add to toolbar
        """
        command_instance = command_class(self.iface, self)
        enabled = getattr(command_instance, 'enabled', True)

        self.add_action(
            command_instance.icon_path(),
            text=self.tr(command_instance.text()),
            add_to_toolbar=add_to_toolbar,
            enabled_flag=enabled,
            callback=command_instance.callback,
            parent=command_instance.parent())

        self.commands.append(command_instance)

    def install_commands(self):
        """Install all commands from the commands directory"""
        modules = import_commands(COMMANDSDIRECTORY)
        self.__add_actions(modules)

    @staticmethod
    @log(print_return=True)
    def __start_qgis_command_tcp_server(command_queue):
        """
        Start TCP server to receive commands from panorama viewer
        
        Args:
            command_queue: Queue for commands
            
        Returns:
            Server port
        """
        server = CommandGetter(command_queue)
        port = server.server_address[1]
        server.start()
        return port

    def __create_proxy_command_line_arguments(self):
        """
        Returns the proxy arguments for the viewer.
        
        Returns:
            List of proxy command line arguments
        """
        from . import proxyutils
        import urllib.request
        
        proxies = urllib.request.getproxies()
        qgis_settings = Settings.getInstance().settings
        proxy_settings = proxyutils.ProxySettings(
            use_proxy_server=qgis_settings[Settings.USE_PROXY_SERVER].value,
            proxy_server_type=qgis_settings[Settings.PROXY_SERVER_TYPE].value,
            proxy_excluded_urls=qgis_settings[Settings.PROXY_EXCLUDED_URLS].value,
            proxy_no_proxy_urls=qgis_settings[Settings.PROXY_NO_PROXY_URLS].value,
            proxy_address=qgis_settings[Settings.PROXY_ADDRESS].value,
            proxy_port=qgis_settings[Settings.PROXY_PORT].value
        )
        return proxyutils.create_proxy_command_line_arguments(proxy_settings, proxies)

    @log(print_return=True)
    def __create_command_line_arguments(self, port, log_port):
        """
        Returns the command line arguments for the viewer process.
        
        Args:
            port: Command server port
            log_port: Log server port
            
        Returns:
            List of command line arguments
        """
        cmd_line_args = [str(port), str(log_port)]
        cmd_line_args.extend(self.__create_proxy_command_line_arguments())
        return cmd_line_args

    @log()
    def start_panorama_viewer(self, port, log_port):
        """
        Start the panorama viewer in another process
        
        Args:
            port: Command server port
            log_port: Log server port
            
        Returns:
            Process handle
        """
        cmd_line_args = self.__create_command_line_arguments(port, log_port)
        logger.debug(f"Starting viewer with args: {cmd_line_args}")
        return start_python_script(
            "panorama_viewer.py", cmd_line_args, not DEBUG_MODE)

    @log()
    def show_viewer(self):
        """Show CEF browser window"""
        if self.panorama_viewer_visible:
            logger.debug("Viewer already visible, not starting again")
            return

        # Command server must be started before CEF is shown
        port = self.__start_qgis_command_tcp_server(self.command_queue)

        # Show CEF
        self.proc = self.start_panorama_viewer(port, self.log_port)

        # Start fetching commands
        self.__fetch_viewer_commands()

    def __fetch_viewer_commands(self):
        """Fetch commands from the panorama viewer periodically"""
        if not self.panorama_viewer_visible:
            logger.debug("Panorama viewer closed, stopping command fetching")
            return
            
        self.command_handler.fetch_commands()
        interval = int(Settings.getInstance().get_timer_interval())
        QTimer.singleShot(interval, self.__fetch_viewer_commands)

    @log()
    def __remove_actions(self):
        """Removes the Qt actions"""
        for action in self.actions:
            self.iface.removePluginMenu(self.tr(u'&StreetSmart'), action)
            self.iface.removeToolBarIcon(action)

    @log()
    def hide_viewer(self):
        """Stops the panorama viewer."""
        if self.panorama_viewer_visible:
            self.send_to_viewer("stopviewer")

    @log()
    def unload(self):
        """Remove the plugin from QGIS"""
        # Hide the viewer
        self.hide_viewer()

        # Stop log service
        self.__kill_process(self.log_pid)

        # Remove actions and toolbar
        self.__remove_actions()
        del self.toolbar

        # Shutdown logging
        logging.info("StreetSmart plugin unloaded")
        logging.shutdown()

    @staticmethod
    @log(print_args=True)
    def __kill_process(pid):
        """
        Kills a process by PID
        
        Args:
            pid: Process ID
        """
        if pid:
            try:
                os.kill(pid, signal.SIGTERM)
                logger.debug(f"Process {pid} terminated")
            except (ProcessLookupError, PermissionError) as e:
                logger.error(f"Failed to kill process {pid}: {str(e)}")


def import_commands(directory: str) -> typing.List:
    """
    Imports all command modules under a given directory.
    
    Args:
        directory: Directory containing command modules
        
    Returns:
        List of tuples (package_parent, module)
    """
    old_cwd = os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    result = []

    try:
        # Get list of Python files in the commands directory
        possible_commands = [f for f in os.listdir(directory) 
                            if f.endswith(".py") and f != "__init__.py"]
        
        for command_file in possible_commands:
            logger.debug(f"Importing {command_file}")
            filename, _ = os.path.splitext(command_file)
            modulename = f"{directory}.{filename}"
            
            # Handle package name for proper import
            if __name__ != '__main__':
                pkg = ".".join(__name__.split('.')[:-1])
                if pkg:
                    modulename = f"{pkg}.{modulename}"
                    pkg_parent = f"{pkg}.{directory}"
                else:
                    pkg_parent = directory
            else:
                pkg_parent = directory
                
            try:
                command_module = importlib.import_module(modulename)
                result.append((pkg_parent, command_module))
            except ImportError as e:
                logger.error(f"Failed to import {modulename}: {str(e)}")
    except Exception as e:
        logger.error(f"Error importing commands: {str(e)}")
    finally:
        # Restore original working directory
        os.chdir(old_cwd)
        
    return result